@@ -2,18 +2,17 @@
 # this repository contains the full copyright notices and license terms.
 
 from trytond.pool import Pool
-from .work import *
-from .timesheet import *
-from .invoice import *
+
+from . import invoice, project, timesheet
 
 
 def register():
     Pool.register(
-        Work,
-        WorkInvoicedProgress,
-        TimesheetLine,
-        InvoiceLine,
+        project.Work,
+        project.WorkInvoicedProgress,
+        timesheet.Line,
+        invoice.InvoiceLine,
         module='project_invoice', type_='model')
     Pool.register(
-        OpenInvoice,
+        project.OpenInvoice,
         module='project_invoice', type_='wizard')
