@@ -2,30 +2,32 @@
 # this repository contains the full copyright notices and license terms.
 
 
-from itertools import groupby
+import datetime
 from collections import defaultdict
 from decimal import Decimal
-import datetime
+from itertools import groupby
 
 from sql import Null
 from sql.aggregate import Sum
+from sql.conditionals import NullIf
 from sql.functions import Extract
 from sql.operators import Concat
-from sql.conditionals import NullIf
 
 from trytond.i18n import gettext
 from trytond.model import ModelSQL, ModelView, fields
-from trytond.pool import PoolMeta
-from trytond.pyson import Eval, Bool, If, Id, PYSONEncoder
-from trytond.pool import Pool
+from trytond.model.exceptions import AccessError
+from trytond.modules.currency.fields import Monetary
+from trytond.pool import Pool, PoolMeta
+from trytond.pyson import Bool, Eval, Id, If
+from trytond.tools import grouped_slice, reduce_ids
 from trytond.transaction import Transaction
-from trytond.wizard import Wizard, StateAction
-from trytond.tools import reduce_ids, grouped_slice
+from trytond.wizard import StateAction, Wizard
 
 from .exceptions import InvoicingError
 
 
 class Effort:
+    __slots__ = ()
 
     @classmethod
     def __setup__(cls):
@@ -38,7 +40,7 @@ class Effort:
         quantities = {}
         for work in works:
             if (work.progress == 1
-                    and work.list_price
+                    and work.invoice_unit_price
                     and not work.invoice_line):
                 if work.price_list_hour:
                     quantity = work.effort_hours
@@ -90,7 +92,7 @@ class Effort:
 
 
 class Progress:
-
+    __slots__ = ()
     invoiced_progress = fields.One2Many('project.work.invoiced_progress',
         'work', 'Invoiced Progress', readonly=True)
 
@@ -114,7 +116,7 @@ class Progress:
         for sub_works in grouped_slice(works):
             sub_works = list(sub_works)
             where = reduce_ids(
-                table.id, [x.id for x in sub_works if x.list_price])
+                table.id, [x.id for x in sub_works if x.invoice_unit_price])
             cursor.execute(*table.join(progress,
                     condition=progress.work == table.id
                     ).select(table.id, Sum(progress.progress),
@@ -126,7 +128,7 @@ class Progress:
                 delta = (
                     (work.progress or 0)
                     - invoiced_progress.get(work.id, 0.0))
-                if work.list_price and delta > 0:
+                if work.invoice_unit_price and delta > 0:
                     quantity = delta
                     if work.price_list_hour:
                         quantity *= work.effort_hours
@@ -216,7 +218,7 @@ class Progress:
 
 
 class Timesheet:
-
+    __slots__ = ()
     project_invoice_timesheet_up_to = fields.Date(
         "Invoice up to",
         states={
@@ -240,7 +242,7 @@ class Timesheet:
                 ()),
             ]
         if 'invoice_method' not in cls.product.depends:
-            cls.product.depends.append('invoice_method')
+            cls.product.depends.add('invoice_method')
 
     @fields.depends('type', 'project_invoice_timesheet_up_to',
         'parent', '_parent_parent.invoice_timesheet_up_to')
@@ -287,7 +289,7 @@ class Timesheet:
         quantities = {}
         for work in works:
             duration = durations[work.id]
-            if work.list_price:
+            if work.invoice_unit_price:
                 hours = duration.total_seconds() / 60 / 60
                 if work.unit_to_invoice:
                     hours = work.unit_to_invoice.round(hours)
@@ -377,19 +379,19 @@ class Work(Effort, Progress, Timesheet, metaclass=PoolMeta):
         'on_change_with_invoice_method')
     quantity_to_invoice = fields.Function(
         fields.Float("Quantity to Invoice"), '_get_invoice_values')
-    amount_to_invoice = fields.Function(fields.Numeric("Amount to Invoice",
-            digits=(16, Eval('currency_digits', 2)),
+    amount_to_invoice = fields.Function(Monetary(
+            "Amount to Invoice", currency='currency', digits='currency',
             states={
                 'invisible': Eval('invoice_method') == 'manual',
                 },
-            depends=['currency_digits', 'invoice_method']),
+            depends=['invoice_method']),
         'get_total')
-    invoiced_amount = fields.Function(fields.Numeric('Invoiced Amount',
-            digits=(16, Eval('currency_digits', 2)),
+    invoiced_amount = fields.Function(Monetary(
+            "Invoiced Amount", currency='currency', digits='currency',
             states={
                 'invisible': Eval('invoice_method') == 'manual',
                 },
-            depends=['currency_digits', 'invoice_method']),
+            depends=['invoice_method']),
         'get_total')
     invoice_line = fields.Many2One('account.invoice.line', 'Invoice Line',
         readonly=True)
@@ -421,6 +423,27 @@ class Work(Effort, Progress, Timesheet, metaclass=PoolMeta):
         default.setdefault('invoice_line', None)
         return super(Work, cls).copy(records, default=default)
 
+    @classmethod
+    def write(cls, *args):
+        actions = iter(args)
+        for works, values in zip(actions, actions):
+            if ('effort_duration' in values
+                    and any(w.invoice_line for w in works)):
+                work = next((w for w in works if w.invoice_line))
+                raise AccessError(gettext(
+                        'project_invoice.msg_invoiced_work_modify_effort',
+                        work=work.rec_name))
+        super().write(*args)
+
+    @classmethod
+    def delete(cls, works):
+        if any(w.invoice_line for w in works):
+            work = next((w for w in works if w.invoice_line))
+            raise AccessError(gettext(
+                    'project_invoice.msg_invoiced_work_delete',
+                    work=work.rec_name))
+        super().delete(works)
+
     @classmethod
     def get_invoice_methods(cls):
         field = 'project_invoice_method'
@@ -446,7 +469,7 @@ class Work(Effort, Progress, Timesheet, metaclass=PoolMeta):
 
     @classmethod
     def _get_amount_to_invoice(cls, works):
-        amounts = {}
+        amounts = defaultdict(Decimal)
         for work in works:
             amounts[work.id] = work.company.currency.round(
                 (work.invoice_unit_price or 0)
@@ -464,7 +487,7 @@ class Work(Effort, Progress, Timesheet, metaclass=PoolMeta):
     @classmethod
     def _get_invoice_values(cls, works, name):
         default = getattr(cls, 'default_%s' % name)
-        amounts = dict.fromkeys((w.id for w in works), default())
+        amounts = defaultdict(default)
         method2works = defaultdict(list)
         for work in works:
             method2works[work.invoice_method].append(work)
@@ -507,13 +530,11 @@ class Work(Effort, Progress, Timesheet, metaclass=PoolMeta):
                 for line in lines:
                     for origin in line['origins']:
                         origins[origin.__class__].append(origin)
-                # TODO: remove when _check_access ignores record rule
-                with Transaction().set_user(0):
-                    for klass, records in origins.items():
-                        klass.save(records)  # Store first new origins
-                        klass.write(records, {
-                                'invoice_line': invoice_line.id,
-                                })
+                for klass, records in origins.items():
+                    klass.save(records)  # Store first new origins
+                    klass.write(records, {
+                            'invoice_line': invoice_line.id,
+                            })
         Invoice.update_taxes(invoices)
 
     def _get_invoice(self):
@@ -558,54 +579,30 @@ class Work(Effort, Progress, Timesheet, metaclass=PoolMeta):
         "Return a invoice line for the lines"
         pool = Pool()
         InvoiceLine = pool.get('account.invoice.line')
-        AccountConfiguration = pool.get('account.configuration')
-        account_config = AccountConfiguration(1)
 
         quantity = sum(l['quantity'] for l in lines)
         product = key['product']
 
-        invoice_line = InvoiceLine()
+        invoice_line = InvoiceLine(invoice=invoice)
+        invoice_line.on_change_invoice()
         invoice_line.type = 'line'
         invoice_line.description = key['description']
-        if product:
-            invoice_line.account = product.account_revenue_used
-            if not invoice_line.account:
+        invoice_line.unit_price = key['unit_price']
+        invoice_line.quantity = quantity
+        invoice_line.unit = key['unit']
+        invoice_line.product = product
+        invoice_line.on_change_product()
+        if not getattr(invoice_line, 'account', None):
+            if invoice_line.product:
                 raise InvoicingError(
                     gettext(
                         'project_invoice.msg_product_missing_account_revenue',
                         work=self.rec_name,
-                        product=product.rec_name))
-        else:
-            invoice_line.account = account_config.get_multivalue(
-                'default_category_account_revenue', company=invoice.company.id)
-            if not invoice_line.account:
+                        product=invoice_line.product.rec_name))
+            else:
                 raise InvoicingError(
                     gettext('project_invoice.msg_missing_account_revenue',
                         work=self.rec_name))
-        invoice_line.product = product
-        invoice_line.unit_price = key['unit_price']
-        invoice_line.quantity = quantity
-        invoice_line.unit = key['unit']
-        invoice_line.currency = invoice.currency
-        invoice_line.company = invoice.company
-
-        taxes = set()
-        pattern = invoice_line._get_tax_rule_pattern()
-        party = invoice.party
-        original_taxes = (
-            product.customer_taxes_used if product else invoice.account.taxes)
-        for tax in original_taxes:
-            if party.customer_tax_rule:
-                tax_ids = party.customer_tax_rule.apply(tax, pattern)
-                if tax_ids:
-                    taxes.update(tax_ids)
-                continue
-            taxes.add(tax.id)
-        if party.customer_tax_rule:
-            tax_ids = party.customer_tax_rule.apply(None, pattern)
-            if tax_ids:
-                taxes.update(tax_ids)
-        invoice_line.taxes = taxes
         return invoice_line
 
     def _test_group_invoice(self):
@@ -666,8 +663,7 @@ class Work(Effort, Progress, Timesheet, metaclass=PoolMeta):
 class WorkInvoicedProgress(ModelView, ModelSQL):
     'Work Invoiced Progress'
     __name__ = 'project.work.invoiced_progress'
-    work = fields.Many2One('project.work', 'Work', ondelete='RESTRICT',
-        select=True)
+    work = fields.Many2One('project.work', "Work", ondelete='RESTRICT')
     progress = fields.Float('Progress', required=True,
         domain=[
             ('progress', '>=', 0),
@@ -721,8 +717,4 @@ class OpenInvoice(Wizard):
             if work.invoiced_progress:
                 for progress in work.invoiced_progress:
                     invoice_ids.add(progress.invoice_line.invoice.id)
-        encoder = PYSONEncoder()
-        action['pyson_domain'] = encoder.encode(
-            [('id', 'in', list(invoice_ids))])
-        action['pyson_search_value'] = encoder.encode([])
-        return action, {}
+        return action, {'res_id': list(invoice_ids)}
